@@ -1,7 +1,7 @@
 #!/usr/bin/env python3
 # -*- coding: utf-8 -*-
 '''
-@author: drunsinn, TobFleischi, Klartext
+@author: drunsinn, TobFleischi, Klartext and GuggiJr
 '''
 
 import datetime
@@ -13,6 +13,7 @@ import string
 import freetype
 import numpy as np
 from scipy.special import binom
+from HersheyFonts import HersheyFonts
 
 class Point():
     def __init__(self, x, y):
@@ -28,7 +29,7 @@ class Point():
 
 class Type2NC(object):
 
-    def __init__(self, output_folder, target_height, step_size, unicode_numbers=None):
+    def __init__(self, output_folder, target_height, step_size, unicode_numbers=None, create_empty_labels=False):
         self._log = logging.getLogger("Type2NC")
         self.__output_folder = output_folder.resolve(strict=True)
         self.__target_height = target_height
@@ -59,16 +60,22 @@ class Type2NC(object):
         else:
             self.__characters.extend(unicode_numbers)
         
+        self.__create_empty_labels = create_empty_labels
+        
         self._log.debug("using folder '%s', step size '%f'", self.__output_folder, self.__step_size)
     
-    def convert(self, font_file):
+    def convert_freetype(self, font_file):
         self._log.debug("running for font file '%s' for %d characters", font_file.resolve(strict=True), len(self.__characters))
 
         font_face = freetype.Face(str(font_file.resolve(strict=True)))
         self._log.info("Font: %s, Style: %s, Format: %s", font_face.family_name.decode("utf-8"), font_face.style_name.decode("utf-8"), font_face.get_format().decode("utf-8"))
         #font_face.set_char_size(height=self.__char_size)
-        font_face.set_char_size(height=self.__char_size_pt * self.__char_size_dpi, hres=self.__char_size_dpi, vres=self.__char_size_dpi)
-        
+        try:
+            font_face.set_char_size(height=self.__char_size_pt * self.__char_size_dpi, hres=self.__char_size_dpi, vres=self.__char_size_dpi)
+        except freetype.ft_errors.FT_Exception as e:
+            self._log.error("Error while setting size for font %s %s : %s", font_face.family_name.decode("utf-8"), font_face.style_name.decode("utf-8"), e)
+            return        
+
         self._log.info("Font BBox (min:max) X:(%d:%d) Y:(%d:%d)", font_face.bbox.xMax, font_face.bbox.xMin, font_face.bbox.yMax, font_face.bbox.yMin)
         scale_factor = self.__target_height / (abs(font_face.bbox.yMax) + abs(font_face.bbox.yMin))
         self._log.info("set scaling factor to %f", scale_factor)
@@ -96,7 +103,12 @@ class Type2NC(object):
                     try:
                         font_face.load_char(chr(char_code))
                     except freetype.ft_errors.FT_Exception as e:
-                        self._log.error("character '%s' at index %d could not be loaded, skip. Error: %s", chr(char_code), char_index, e)
+                        if self.__create_empty_labels:
+                            self._log.error("character '%s' at index %d could not be loaded, replace with empty label. Error: %s", chr(char_code), char_index, e)
+                            self._log.debug("character '%s' was selected but is not available, add empty label", chr(char_code))
+                            ncfp.writelines(self._create_empty_font_label(chr(char_code), x_advance=0))
+                        else:
+                            self._log.error("character '%s' at index %d could not be loaded, skip. Error: %s", chr(char_code), char_index, e)
                         break
 
                     c_glyph_slot = font_face.glyph
@@ -163,7 +175,11 @@ class Type2NC(object):
                     ncfp.writelines(self._creat_font_label(chr(char_code), contour_paths, x_advance, scale_factor))
                     
                 else:
-                    self._log.debug("character '%s' was selected but is not available, skipping", chr(char_code))
+                    if self.__create_empty_labels:
+                        self._log.debug("character '%s' was selected but is not available, add empty label", chr(char_code))
+                        ncfp.writelines(self._create_empty_font_label(chr(char_code), x_advance=0))
+                    else:
+                        self._log.debug("character '%s' was selected but is not available, skipping", chr(char_code))
 
             with open(pathlib.Path.cwd().joinpath("templates", "pgm_foot_template.H")) as template_file:
                     ncfp.writelines(template_file)
@@ -172,12 +188,71 @@ class Type2NC(object):
 
         self._log.info("finished writing %s", nc_file_name)
 
-    def _build_char_string(self, range_list):
-        characters = list()
-        print(range_list)
-        for char_range in range_list:
-            characters.extend(char_range)
-        return characters
+    def convert_hershey(self, font_file):
+        hshf = HersheyFonts()
+        if isinstance(font_file, pathlib.Path):
+            self._log.debug("running for hershey font file '%s' for %d characters", font_file.resolve(strict=True), len(self.__characters))
+            hshf.load_font_file(font_file)
+            nc_file_name = font_file.name.replace(font_file.suffix, '')
+        else:
+            self._log.debug("running for hershey default font '%s' for %d characters", font_file, len(self.__characters))
+            hshf.load_default_font(font_file)
+            nc_file_name = font_file
+
+        hshf.normalize_rendering(self.__target_height * 0.8)
+
+        nc_file_name = nc_file_name.replace(' ', '_')
+        nc_file_path = self.__output_folder.joinpath(nc_file_name + '.H')
+
+        scale_factor = 1.0
+
+        with open(nc_file_path, 'w') as ncfp:
+            ncfp.write("BEGIN PGM {0:s} MM\n".format(nc_file_name.upper()))
+                    
+            ncfp.write(";\n; Font PGM generated by type2nc\n")
+            ncfp.write("; Hershey Font File: {0:s}\n".format(nc_file_name))
+            ncfp.write("; Generated: {:%Y-%m-%d %H:%M:%S}\n".format(datetime.datetime.today()))
+            ncfp.write("; Number of characters: {0:d}\n;\n".format(len(self.__characters)))
+
+            with open(pathlib.Path.cwd().joinpath("templates", "pgm_head_template.H")) as template_file:
+                ncfp.writelines(template_file)
+
+            for char_code in self.__characters:
+                
+                count = 0
+                contour_paths = list()
+                for glyph in hshf.glyphs_for_text(chr(char_code)):
+                    if count > 0:
+                        raise Exception
+                    count += 1
+
+                    x_advance = glyph.char_width + 1.0
+
+                    self._log.debug("character '%s' advance X:%d", chr(char_code), x_advance)
+
+                    for stroke in glyph.strokes:
+                        path_points = list()
+                        for c_pair in stroke:
+                            x_coor = c_pair[0] - glyph.left_offset
+                            y_coor = -1 * c_pair[1]
+                            path_points.append(Point(x=x_coor, y=y_coor))
+                        contour_paths.append(path_points)
+                    self._log.debug("created %d paths for character", len(contour_paths))
+                    
+                    ncfp.writelines(self._creat_font_label(chr(char_code), contour_paths, x_advance, scale_factor=scale_factor))
+
+                if len(contour_paths) == 0:
+                    if self.__create_empty_labels:
+                        self._log.debug("character '%s' was selected but is not available, add empty label", chr(char_code))
+                        ncfp.writelines(self._create_empty_font_label(chr(char_code), x_advance=0))
+                    else:
+                        self._log.debug("character '%s' was selected but is not available, skipping", chr(char_code))
+
+            with open(pathlib.Path.cwd().joinpath("templates", "pgm_foot_template.H")) as template_file:
+                    ncfp.writelines(template_file)
+
+            ncfp.write("END PGM {0:s} MM".format(nc_file_name.upper()))
+
 
     def _point_on_curve(self, point_list, distance_factor):
         """Get x and y coordinate for point along a bezier curve relative to
@@ -228,6 +303,16 @@ class Type2NC(object):
         char_lines.append("LBL 0\n;\n")
         return char_lines
 
+    def _create_empty_font_label(self, char_str, x_advance):
+        self._log.debug("created label for empty char %s", char_str)
+        char_lines = list()
+        char_lines.append("* -   Unicode Hex:0x{0:04x}\n".format(ord(char_str)))
+        char_lines.append("LBL \"0x{0:04x}\"\n".format(ord(char_str)))
+        char_lines.append("LBL \"0x{0:04x}_X-Advance\"\n".format(ord(char_str)))
+        char_lines.append("QL20 = {0:+f} ; X-Advance\n".format(x_advance))
+        char_lines.append("LBL 0\n;\n")
+        return char_lines
+
     def _translate_label_name(self, char_str):
         label_map = {
             '!': 'exclamation_mark',
@@ -311,8 +396,8 @@ class Type2NC_UI:
 
         self._window_root = root
         root.title("Type2NC UI")
-        width=650
-        height=260
+        width=670
+        height=300
         screenwidth = root.winfo_screenwidth()
         screenheight = root.winfo_screenheight()
         alignstr = '%dx%d+%d+%d' % (width, height, (screenwidth - width) / 2, (screenheight - height) / 2)
@@ -336,9 +421,22 @@ class Type2NC_UI:
         self.lbl_font_filename["font"] = ft
         self.lbl_font_filename["justify"] = "center"
         self.lbl_font_filename["text"] = " "
+        self.lbl_font_filename["anchor"] = "w"
         self.lbl_font_filename.place(x=200, y=current_y, width=180, height=component_height)
         current_y += delta_y
 
+        self.include_hershey = tk.IntVar()
+        self.chk_hershey = tk.Checkbutton(self._window_root)
+        self.chk_hershey["font"] = ft
+        self.chk_hershey["justify"] = "left"
+        self.chk_hershey["text"] = self.gt_("Include Default Hershey Fonts")
+        self.chk_hershey.place(x=20, y=current_y, width=180, height=component_height)
+        self.chk_hershey["onvalue"] = 1
+        self.chk_hershey["offvalue"] = 0
+        self.chk_hershey["variable"] = self.include_hershey
+        self.chk_hershey.deselect()
+        current_y += delta_y
+
         self.btn_select_folder = tk.Button(self._window_root)
         self.btn_select_folder["font"] = ft
         self.btn_select_folder["justify"] = "center"
@@ -350,6 +448,7 @@ class Type2NC_UI:
         self.lbl_output_path["font"] = ft
         self.lbl_output_path["justify"] = "center"
         self.lbl_output_path["text"] = " "
+        self.lbl_output_path["anchor"] = "w"
         self.lbl_output_path.place(x=200, y=current_y, width=180, height=component_height)
         current_y += delta_y
 
@@ -364,6 +463,7 @@ class Type2NC_UI:
         self.lbl_step_size["font"] = ft
         self.lbl_step_size["justify"] = "center"
         self.lbl_step_size["text"] = " "
+        self.lbl_step_size["anchor"] = "w"
         self.lbl_step_size.place(x=200, y=current_y, width=180, height=component_height)
         current_y += delta_y
 
@@ -379,6 +479,18 @@ class Type2NC_UI:
         self.chk_generate_demos.select()
         current_y += delta_y
 
+        self.include_empty_lbl = tk.IntVar()
+        self.chk_empty_lbl = tk.Checkbutton(self._window_root)
+        self.chk_empty_lbl["font"] = ft
+        self.chk_empty_lbl["justify"] = "left"
+        self.chk_empty_lbl["text"] = self.gt_("Include Empty Label")
+        self.chk_empty_lbl.place(x=20, y=current_y, width=150, height=component_height)
+        self.chk_empty_lbl["onvalue"] = 1
+        self.chk_empty_lbl["offvalue"] = 0
+        self.chk_empty_lbl["variable"] = self.include_empty_lbl
+        self.chk_empty_lbl.select()
+        current_y += delta_y
+
         self.btn_generate_nc = tk.Button(self._window_root)
         self.btn_generate_nc["font"] = ft
         self.btn_generate_nc["justify"] = "center"
@@ -400,7 +512,8 @@ class Type2NC_UI:
         self.chk_select_basic["font"] = ft
         self.chk_select_basic["justify"] = "left"
         self.chk_select_basic["text"] = self.gt_("ASCII 0x20-0x7E and Latin1 0x80-0xFF")
-        self.chk_select_basic.place(x=390, y=current_y, width=240, height=component_height)
+        self.chk_select_basic["anchor"] = "w"
+        self.chk_select_basic.place(x=390, y=current_y, width=260, height=component_height)
         self.chk_select_basic["onvalue"] = 1
         self.chk_select_basic["offvalue"] = 0
         self.chk_select_basic["variable"] = self.select_basics
@@ -412,7 +525,8 @@ class Type2NC_UI:
         self.chk_select_punctuation["font"] = ft
         self.chk_select_punctuation["justify"] = "left"
         self.chk_select_punctuation["text"] = self.gt_("General Punctuation 0x2000-0x206F")
-        self.chk_select_punctuation.place(x=390, y=current_y, width=240, height=component_height)
+        self.chk_select_punctuation["anchor"] = "w"
+        self.chk_select_punctuation.place(x=390, y=current_y, width=260, height=component_height)
         self.chk_select_punctuation["onvalue"] = 1
         self.chk_select_punctuation["offvalue"] = 0
         self.chk_select_punctuation["variable"] = self.select_punctuation
@@ -423,8 +537,9 @@ class Type2NC_UI:
         self.chk_select_ipa = tk.Checkbutton(self._window_root)
         self.chk_select_ipa["font"] = ft
         self.chk_select_ipa["justify"] = "left"
-        self.chk_select_ipa["text"] = self.gt_("IPA Extention 0x0250-0x02AF")
-        self.chk_select_ipa.place(x=390, y=current_y, width=240, height=component_height)
+        self.chk_select_ipa["text"] = self.gt_("IPA Extension 0x0250-0x02AF")
+        self.chk_select_ipa["anchor"] = "w"
+        self.chk_select_ipa.place(x=390, y=current_y, width=260, height=component_height)
         self.chk_select_ipa["onvalue"] = 1
         self.chk_select_ipa["offvalue"] = 0
         self.chk_select_ipa["variable"] = self.select_ipa
@@ -436,7 +551,8 @@ class Type2NC_UI:
         self.chk_select_symbols["font"] = ft
         self.chk_select_symbols["justify"] = "left"
         self.chk_select_symbols["text"] = self.gt_("Symbols 0x2190-0x23FF & 0x2600-0x27BF")
-        self.chk_select_symbols.place(x=390, y=current_y, width=240, height=component_height)
+        self.chk_select_symbols["anchor"] = "w"
+        self.chk_select_symbols.place(x=390, y=current_y, width=260, height=component_height)
         self.chk_select_symbols["onvalue"] = 1
         self.chk_select_symbols["offvalue"] = 0
         self.chk_select_symbols["variable"] = self.select_symbols
@@ -448,7 +564,8 @@ class Type2NC_UI:
         self.chk_select_lang["font"] = ft
         self.chk_select_lang["justify"] = "left"
         self.chk_select_lang["text"] = self.gt_("Lang 0x0370-0x077F & 0x4E00-0x9FFF")
-        self.chk_select_lang.place(x=390, y=current_y, width=240, height=component_height)
+        self.chk_select_lang["anchor"] = "w"
+        self.chk_select_lang.place(x=390, y=current_y, width=260, height=component_height)
         self.chk_select_lang["onvalue"] = 1
         self.chk_select_lang["offvalue"] = 0
         self.chk_select_lang["variable"] = self.select_add_lang
@@ -464,8 +581,7 @@ class Type2NC_UI:
     def btn_select_font_command(self):
         self._log.debug("select font button pressed")
         
-
-        file_types = [('Font', '*.ttf *.tte *.ttc *.otf *.dfont *.pfb')]
+        file_types = [('Font', '*.ttf *.tte *.ttc *.otf *.dfont *.pfb'), ('Hershey Font', '*.jhf')]
         font_file_list = tkfd.askopenfilename(parent=self._window_root, title=self.gt_("Select Font File"), filetypes=file_types, multiple=1)
         for path in font_file_list:
             self.selected_font_files.append(pathlib.Path(path))
@@ -480,7 +596,6 @@ class Type2NC_UI:
     def btn_select_folder_command(self):
         self._log.debug("select output folder button pressed")
         
-
         selected_folder = tkfd.askdirectory(parent=self._window_root, title=self.gt_("Select Destination Folder"))
         self.output_path = pathlib.Path(selected_folder)
 
@@ -495,7 +610,6 @@ class Type2NC_UI:
     def btn_select_step_command(self):
         self._log.debug("select step size button pressed")
         
-
         selected_step = tksd.askfloat(parent=self._window_root, title=self.gt_("Step Size"), prompt=self.gt_("Step Size between 0.001 (very fine) and 0.2 (very coarse) for converting Splines"), initialvalue=0.05, minvalue=0.001, maxvalue=0.2)
         if selected_step is None:
             self.step_size = None
@@ -509,8 +623,7 @@ class Type2NC_UI:
     def btn_generate_nc_command(self):
         self._log.debug("generate nc files button pressed")
         
-
-        if len(self.selected_font_files) == 0:
+        if len(self.selected_font_files) == 0 and self.include_hershey.get() == 0:
             tkmb.showwarning(parent=self._window_root, title=self.gt_("Missing parameter"), message=self.gt_("No font files selected"))
             self._log.debug("list of selected font files is empty!")
             return
@@ -519,7 +632,7 @@ class Type2NC_UI:
             self._log.debug("no output path selected!")
             return
         if self.step_size is None:
-            tkmb.showwarning(parent=self._window_root, title=self.gt_("Missing parameter"), message=-("No step size selected"))
+            tkmb.showwarning(parent=self._window_root, title=self.gt_("Missing parameter"), message=self.gt_("No step size selected"))
             self._log.debug("no step size selected selected!")
             return
         
@@ -553,11 +666,24 @@ class Type2NC_UI:
             character_list.extend(list(range(0x4E00, 0x9FFF + 1))) # CJK_UNIFIED_IDEOGRAPHS_PART
 
         self.pb_progress['value'] = 0
-        conv = Type2NC(output_folder=self.output_path, target_height=10, step_size=self.step_size, unicode_numbers=character_list)
+
+        include_empty = False
+        if self.include_empty_lbl.get() > 0:
+            include_empty = True
+
+        conv = Type2NC(output_folder=self.output_path, target_height=10, step_size=self.step_size, unicode_numbers=character_list, create_empty_labels=include_empty)
         self._log.debug("start processing %d font files", len(self.selected_font_files))
+        
+        if self.include_hershey.get() > 0:
+            for hershey_font in HersheyFonts().default_font_names:
+                conv.convert_hershey(hershey_font)
+        
         for i, ff in enumerate(self.selected_font_files):
             if ff.is_file():
-                conv.convert(ff)
+                if ff.suffix.lower() in ".jhf":
+                    conv.convert_hershey(ff)
+                else:
+                    conv.convert_freetype(ff)
             self.pb_progress['value'] = (i + 1) * (100 / len(self.selected_font_files))
             self.pb_progress.update()
 
@@ -574,24 +700,38 @@ class Type2NC_UI:
 if __name__ == "__main__":
     logging.basicConfig(level=logging.INFO)
     logger = logging.getLogger('main')
-    logger.debug("startup")
+    logger.info("startup")
+    log_level_map = {'critical': logging.CRITICAL, 'error': logging.ERROR,
+                     'warn': logging.WARNING, 'warning': logging.WARNING,
+                     'info': logging.INFO, 'debug': logging.DEBUG}
 
     cmdl_parser = argparse.ArgumentParser(description="Create Klartext NC code from font files. If no options are given, start gui")
     cmdl_parser.add_argument("-i", "--input", metavar="font input file", nargs='+', type=pathlib.Path, help="path of one or more font files")
     cmdl_parser.add_argument("-o", "--output", metavar="output folder", type=pathlib.Path, help="path to the output folder where klartext files are generated")
     cmdl_parser.add_argument("-s", "--step_size", metavar="step size", type=float, default=0.05, required=False, help="step size for converting curves to line segmenst: between 0.001 (very fine) and 0.2 (very coarse)")
     cmdl_parser.add_argument("-d", "--create_demos", action="store_true", default=False, required=False, help="if set, demo output will use cycle 225 for definition of parameters")
+    cmdl_parser.add_argument("-e", "--create_empty_label", action="store_true", default=False, required=False, help="if set, create label for each selected character, even if it is not defined in the font. Stops errors because of missing label definition")
+    cmdl_parser.add_argument("-l", "--log", metavar="logging level", default="warning", help="set logging level to critical, error, warn/warning, info or debug. Default is info")
+
     arguments = cmdl_parser.parse_args()
 
+    selected_level = log_level_map.get(arguments.log.lower())
+
+    if selected_level is None:
+        raise Exception("the given log level {:s} cant be mapped to an existing log level. It must be one of critical, error, warn/warning, info or debug")
+    logging.basicConfig(level=selected_level)
+
     if arguments.input is not None:
         if not arguments.output.is_dir():
             logger.error("the output path '%s' is not a folder", arguments.output)
             
-        conv = Type2NC(output_folder=arguments.output, target_height=10, step_size=arguments.step_size)
+        conv = Type2NC(output_folder=arguments.output, target_height=10, step_size=arguments.step_size, create_empty_labels=arguments.create_empty_label)
         for ff in arguments.input:
             if ff.is_file():
-                
-                conv.convert(ff)
+                if ff.suffix.lower() in ".jhf":
+                    conv.convert_hershey(ff)
+                else:
+                    conv.convert_freetype(ff)
         
         if arguments.create_demos:
             conv.generate_demo_files(arguments.input)
